#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <curl/curl.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/epoll.h>
#include <errno.h>
#include <vector>
#include <string>
#include <map>
#include <algorithm>

#include "metacube.h"
#include "server.h"
#include "mutexlock.h"
#include "parse.h"
#include "state.pb.h"

using namespace std;

Client::Client(int sock)
	: sock(sock),
	  state(Client::READING_REQUEST),
	  header_or_error_bytes_sent(0),
	  bytes_sent(0)
{
	request.reserve(1024);
}
	
Client::Client(const ClientProto &serialized)
	: sock(serialized.sock()),
	  state(State(serialized.state())),
	  request(serialized.request()),
	  stream_id(serialized.stream_id()),
	  header_or_error(serialized.header_or_error()),
	  header_or_error_bytes_sent(serialized.header_or_error_bytes_sent()),
	  bytes_sent(serialized.bytes_sent())
{
}

ClientProto Client::serialize() const
{
	ClientProto serialized;
	serialized.set_sock(sock);
	serialized.set_state(state);
	serialized.set_request(request);
	serialized.set_stream_id(stream_id);
	serialized.set_header_or_error(header_or_error);
	serialized.set_header_or_error_bytes_sent(serialized.header_or_error_bytes_sent());
	serialized.set_bytes_sent(bytes_sent);
	return serialized;
}

Stream::Stream(const string &stream_id)
	: stream_id(stream_id),
	  data(new char[BACKLOG_SIZE]),
	  data_size(0)
{
	memset(data, 0, BACKLOG_SIZE);
}

Stream::~Stream()
{
	delete[] data;
}

Stream::Stream(const StreamProto &serialized)
	: stream_id(serialized.stream_id()),
	  header(serialized.header()),
	  data(new char[BACKLOG_SIZE]),
	  data_size(serialized.data_size())
{
	assert(serialized.data().size() == BACKLOG_SIZE);
	memcpy(data, serialized.data().data(), BACKLOG_SIZE);
}

StreamProto Stream::serialize() const
{
	StreamProto serialized;
	serialized.set_header(header);
	serialized.set_data(string(data, data + BACKLOG_SIZE));
	serialized.set_data_size(data_size);
	serialized.set_stream_id(stream_id);
	return serialized;
}

Server::Server()
{
	pthread_mutex_init(&mutex, NULL);

	epoll_fd = epoll_create(1024);  // Size argument is ignored.
	if (epoll_fd == -1) {
		perror("epoll_fd");
		exit(1);
	}
}

Server::~Server()
{
	close(epoll_fd);
}

void Server::run()
{
	should_stop = false;
	
	// Joinable is already the default, but it's good to be certain.
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_create(&worker_thread, &attr, Server::do_work_thunk, this);
}
	
void Server::stop()
{
	{
		MutexLock lock(&mutex);
		should_stop = true;
	}

	if (pthread_join(worker_thread, NULL) == -1) {
		perror("pthread_join");
		exit(1);
	}
}

void *Server::do_work_thunk(void *arg)
{
	Server *server = static_cast<Server *>(arg);
	server->do_work();
	return NULL;
}

void Server::do_work()
{
	for ( ;; ) {
		int nfds = epoll_wait(epoll_fd, events, EPOLL_MAX_EVENTS, EPOLL_TIMEOUT_MS);
		if (nfds == -1) {
			perror("epoll_wait");
			exit(1);
		}

		MutexLock lock(&mutex);  // We release the mutex between iterations.
	
		if (should_stop) {
			return;
		}
	
		for (int i = 0; i < nfds; ++i) {
			int fd = events[i].data.fd;
			assert(clients.count(fd) != 0);
			Client *client = &clients[fd];

			if (events[i].events & (EPOLLERR | EPOLLRDHUP | EPOLLHUP)) {
				close_client(client);
				continue;
			}

			process_client(client);
		}
	}
}

CubemapStateProto Server::serialize() const
{
	CubemapStateProto serialized;
	for (map<int, Client>::const_iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		serialized.add_clients()->MergeFrom(client_it->second.serialize());
	}
	for (map<string, Stream *>::const_iterator stream_it = streams.begin();
	     stream_it != streams.end();
	     ++stream_it) {
		serialized.add_streams()->MergeFrom(stream_it->second->serialize());
	}
	return serialized;
}

void Server::add_client(int sock)
{
	MutexLock lock(&mutex);
	clients.insert(make_pair(sock, Client(sock)));

	// Start listening on data from this socket.
	epoll_event ev;
	ev.events = EPOLLIN | EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = sock;
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_ADD)");
		exit(1);
	}
}

void Server::add_client_from_serialized(const ClientProto &client)
{
	MutexLock lock(&mutex);
	clients.insert(make_pair(client.sock(), Client(client)));

	// Start listening on data from this socket.
	epoll_event ev;
	if (client.state() == Client::READING_REQUEST) {
		ev.events = EPOLLIN | EPOLLRDHUP;
	} else {
		// If we don't have more data for this client, we'll be putting it into
		// the sleeping array again soon.
		ev.events = EPOLLOUT | EPOLLRDHUP;
	}
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client.sock();
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, client.sock(), &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_ADD)");
		exit(1);
	}
}

void Server::add_stream(const string &stream_id)
{
	MutexLock lock(&mutex);
	streams.insert(make_pair(stream_id, new Stream(stream_id)));
}

void Server::add_stream_from_serialized(const StreamProto &stream)
{
	MutexLock lock(&mutex);
	streams.insert(make_pair(stream.stream_id(), new Stream(stream)));
}
	
void Server::set_header(const string &stream_id, const string &header)
{
	MutexLock lock(&mutex);
	find_stream(stream_id)->header = header;

	// If there are clients we haven't sent anything to yet, we should give
	// them the header, so push back into the SENDING_HEADER state.
	for (map<int, Client>::iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		Client *client = &client_it->second;
		if (client->state == Client::SENDING_DATA &&
		    client->bytes_sent == 0) {
			construct_header(client);
		}
	}
}
	
void Server::add_data(const string &stream_id, const char *data, size_t bytes)
{
	if (bytes == 0) {
		return;
	}

	MutexLock lock(&mutex);
	Stream *stream = find_stream(stream_id);
	size_t pos = stream->data_size % BACKLOG_SIZE;
	stream->data_size += bytes;

	if (pos + bytes > BACKLOG_SIZE) {
		size_t to_copy = BACKLOG_SIZE - pos;
		memcpy(stream->data + pos, data, to_copy);
		data += to_copy;
		bytes -= to_copy;
		pos = 0;
	}

	memcpy(stream->data + pos, data, bytes);
	wake_up_all_clients();
}
	
void Server::process_client(Client *client)
{
	switch (client->state) {
	case Client::READING_REQUEST: {
		// Try to read more of the request.
		char buf[1024];
		int ret = read(client->sock, buf, sizeof(buf));
		if (ret == -1) {
			perror("read");
			close_client(client);
			return;
		}
		if (ret == 0) {
			// No data? This really means that we were triggered for something else than
			// POLLIN (which suggests a logic error in epoll).
			fprintf(stderr, "WARNING: fd %d returned unexpectedly 0 bytes!\n", client->sock);
			close_client(client);
			return;
		}

		// Guard against overlong requests gobbling up all of our space.
		if (client->request.size() + ret > MAX_CLIENT_REQUEST) {
			fprintf(stderr, "WARNING: fd %d sent overlong request!\n", client->sock);
			close_client(client);
			return;
		}	

		// See if we have \r\n\r\n anywhere in the request. We start three bytes
		// before what we just appended, in case we just got the final character.
		size_t existing_req_bytes = client->request.size();
		client->request.append(string(buf, buf + ret));
	
		size_t start_at = (existing_req_bytes >= 3 ? existing_req_bytes - 3 : 0);
		const char *ptr = reinterpret_cast<char *>(
			memmem(client->request.data() + start_at, client->request.size() - start_at,
			       "\r\n\r\n", 4));
		if (ptr == NULL) {
			// OK, we don't have the entire header yet. Fine; we'll get it later.
			return;
		}

		if (ptr != client->request.data() + client->request.size() - 4) {
			fprintf(stderr, "WARNING: fd %d had junk data after request!\n", client->sock);
			close_client(client);
			return;
		}

		int error_code = parse_request(client);
		if (error_code == 200) {
			construct_header(client);
		} else {
			construct_error(client, error_code);
		}
		break;
	}
	case Client::SENDING_ERROR:
	case Client::SENDING_HEADER: {
		int ret = write(client->sock,
		                client->header_or_error.data() + client->header_or_error_bytes_sent,
		                client->header_or_error.size() - client->header_or_error_bytes_sent);
		if (ret == -1) {
			perror("write");
			close_client(client);
			return;
		}
		
		client->header_or_error_bytes_sent += ret;
		assert(client->header_or_error_bytes_sent <= client->header_or_error.size());

		if (client->header_or_error_bytes_sent < client->header_or_error.size()) {
			// We haven't sent all yet. Fine; we'll do that later.
			return;
		}

		// We're done sending the header or error! Clear it to release some memory.
		client->header_or_error.clear();

		if (client->state == Client::SENDING_ERROR) {
			close_client(client);
		} else {
			// Start sending from the end. In other words, we won't send any of the backlog,
			// but we'll start sending immediately as we get data.
			client->state = Client::SENDING_DATA;
			client->bytes_sent = find_stream(client->stream_id)->data_size;
		}
		break;
	}
	case Client::SENDING_DATA: {
		// See if there's some data we've lost. Ideally, we should drop to a block boundary,
		// but resync will be the mux's problem.
		const Stream &stream = *find_stream(client->stream_id);
		size_t bytes_to_send = stream.data_size - client->bytes_sent;
		if (bytes_to_send > BACKLOG_SIZE) {
			fprintf(stderr, "WARNING: fd %d lost %lld bytes, maybe too slow connection\n",
			        client->sock,
			        (long long int)(bytes_to_send - BACKLOG_SIZE));
			client->bytes_sent = find_stream(client->stream_id)->data_size - BACKLOG_SIZE;
			bytes_to_send = BACKLOG_SIZE;
		}

		// See if we need to split across the circular buffer.
		ssize_t ret;
		if ((client->bytes_sent % BACKLOG_SIZE) + bytes_to_send > BACKLOG_SIZE) {
			size_t bytes_first_part = BACKLOG_SIZE - (client->bytes_sent % BACKLOG_SIZE);

			iovec iov[2];
			iov[0].iov_base = const_cast<char *>(stream.data + (client->bytes_sent % BACKLOG_SIZE));
			iov[0].iov_len = bytes_first_part;

			iov[1].iov_base = const_cast<char *>(stream.data);
			iov[1].iov_len = bytes_to_send - bytes_first_part;

			ret = writev(client->sock, iov, 2);
		} else {
			ret = write(client->sock,
			            stream.data + (client->bytes_sent % BACKLOG_SIZE),
			            bytes_to_send);
		}
		if (ret == -1) {
			perror("write/writev");
			close_client(client);
			return;
		}
		client->bytes_sent += ret;

		if (client->bytes_sent == stream.data_size) {
			// We don't have any more data for this client, so put it to sleep.
			put_client_to_sleep(client);
		}
		break;
	}
	default:
		assert(false);
	}
}

int Server::parse_request(Client *client)
{
	vector<string> lines = split_lines(client->request);
	if (lines.empty()) {
		return 400;  // Bad request (empty).
	}

	vector<string> request_tokens = split_tokens(lines[0]);
	if (request_tokens.size() < 2) {
		return 400;  // Bad request (empty).
	}
	if (request_tokens[0] != "GET") {
		return 400;  // Should maybe be 405 instead?
	}
	if (streams.count(request_tokens[1]) == 0) {
		return 404;  // Not found.
	}

	client->stream_id = request_tokens[1];
	client->request.clear();

	return 200;  // OK!
}

void Server::construct_header(Client *client)
{
	client->header_or_error = "HTTP/1.0 200 OK\r\nContent-type: video/x-flv\r\nCache-Control: no-cache\r\n\r\n" +
		find_stream(client->stream_id)->header;

	// Switch states.
	client->state = Client::SENDING_HEADER;

	epoll_event ev;
	ev.events = EPOLLOUT | EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client->sock;

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}
}
	
void Server::construct_error(Client *client, int error_code)
{
	char error[256];
	snprintf(error, 256, "HTTP/1.0 %d Error\r\nContent-type: text/plain\r\n\r\nSomething went wrong. Sorry.\r\n",
		error_code);
	client->header_or_error = error;

	// Switch states.
	client->state = Client::SENDING_ERROR;

	epoll_event ev;
	ev.events = EPOLLOUT | EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client->sock;

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}
}
	
void Server::close_client(Client *client)
{
	if (epoll_ctl(epoll_fd, EPOLL_CTL_DEL, client->sock, NULL) == -1) {
		perror("epoll_ctl(EPOLL_CTL_DEL)");
		exit(1);
	}

	// This client could be sleeping, so we'll need to fix that. (Argh, O(n).)
	vector<int>::iterator new_end =
		remove(sleeping_clients.begin(), sleeping_clients.end(), client->sock);
	sleeping_clients.erase(new_end, sleeping_clients.end());
	
	// Bye-bye!
	close(client->sock);
	clients.erase(client->sock);
}
	
void Server::put_client_to_sleep(Client *client)
{
	epoll_event ev;
	ev.events = EPOLLRDHUP;
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.fd = client->sock;

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}

	sleeping_clients.push_back(client->sock);
}

void Server::wake_up_all_clients()
{
	for (unsigned i = 0; i < sleeping_clients.size(); ++i) {
		epoll_event ev;
		ev.events = EPOLLOUT | EPOLLRDHUP;
		ev.data.u64 = 0;  // Keep Valgrind happy.
		ev.data.fd = sleeping_clients[i];
		if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, sleeping_clients[i], &ev) == -1) {
			perror("epoll_ctl(EPOLL_CTL_MOD)");
			exit(1);
		}
	}
	sleeping_clients.clear();
}
	
Stream *Server::find_stream(const string &stream_id)
{
	map<string, Stream *>::iterator it = streams.find(stream_id);
	assert(it != streams.end());
	return it->second;
}
